-------------------------------------------------------------------------------
--
-- Text-mode VGA controller for the Digilent Spartan 3 starter board
--
-- Stephen A. Edwards
-- sedwards@cs.columbia.edu
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity s3board_vga is

  port (
    CLK : in std_logic;
    RST : in std_logic;

    -- Video signals
--    VIDOUT_CLK     : out std_logic;
    VIDOUT_RED     : out std_logic;
    VIDOUT_GREEN   : out std_logic;
    VIDOUT_BLUE    : out std_logic;
    VIDOUT_HSYNC : out std_logic;
    VIDOUT_VSYNC : out std_logic);

end s3board_vga;

architecture Behavioral of s3board_vga is

  -- Video parameters

  constant HTOTAL : integer := 800;
  constant HSYNC : integer := 96;
  constant HBACK_PORCH : integer := 48;
  constant HACTIVE : integer := 640;
  constant HFRONT_PORCH : integer := 16;

  constant VTOTAL : integer := 525;
  constant VSYNC : integer := 2;
  constant VBACK_PORCH : integer := 33;
  constant VACTIVE : integer := 480;
  constant VFRONT_PORCH : integer := 10;

  component RAMB16_S9
    port (
      DO   : out std_logic_vector (7 downto 0);
      DOP0  : out std_logic;
      ADDR : in std_logic_vector (10 downto 0);
      CLK  : in std_logic;
      DI   : in std_logic_vector (7 downto 0);
      DIP0  : in std_logic;
      EN   : in std_logic;
      SSR  : in std_logic;
      WE   : in std_logic);
  end component;

  component RAMB16_S36
    port (
      DO   : out std_logic_vector (31 downto 0);
      DOP  : out std_logic_vector(3 downto 0);
      ADDR : in std_logic_vector (8 downto 0);
      CLK  : in std_logic;
      DI   : in std_logic_vector (31 downto 0);
      DIP  : in std_logic_vector(3 downto 0);
      EN   : in std_logic;
      SSR  : in std_logic;
      WE   : in std_logic);
  end component;

  -- Signals for the video controller
  signal Pixel_Clock : std_logic;       -- 25 MHz clock divided from 50 MHz
  signal LoadNShift : std_logic;        -- Shift register control
  signal FontData : std_logic_vector(31 downto 0);   -- Input to shift register
  signal ShiftData : std_logic_vector(31 downto 0);  -- Shift register data
  signal VideoData : std_logic;         -- Serial out ANDed with blanking

  signal Hcount : std_logic_vector(9 downto 0);  -- Horizontal position (0-800)
  signal Vcount : std_logic_vector(9 downto 0);  -- Vertical position (0-524)
  signal HBLANK_N, VBLANK_N : std_logic;   -- Blanking signals
  signal RED, GREEN, BLUE : std_logic;

--  signal FontLoad, LoadChar : std_logic;  -- Font/Character RAM read triggers
  signal FontAddr : std_logic_vector(8 downto 0);
  signal CharAddr : std_logic_vector(10 downto 0);
  signal CharColumn : std_logic_vector(9 downto 0);
  signal CharRow : std_logic_vector(9 downto 0);
  signal Column : std_logic_vector(4 downto 0); -- 0-19
  signal Row : std_logic_vector(3 downto 0);   -- 0-14
  signal EndOfLine, EndOfField : std_logic;
  signal Background, Player, Cursor : std_logic;

  signal DOUTB0 : std_logic_vector(7 downto 0);
  signal DOUTB5 : std_logic_vector(31 downto 0);
  signal DOP0 : std_logic;
  signal DOP5 : std_logic_vector(3 downto 0);

  attribute INIT_00 : string;
  attribute INIT_01 : string;
  attribute INIT_02 : string;
  attribute INIT_03 : string;
  attribute INIT_04 : string;
  attribute INIT_05 : string;
  attribute INIT_06 : string;
  attribute INIT_07 : string;
  attribute INIT_08 : string;
  attribute INIT_09 : string;
  attribute INIT_0a : string;
  attribute INIT_0b : string;
  attribute INIT_0c : string;
  attribute INIT_0d : string;
  attribute INIT_0e : string;
  attribute INIT_0f : string;
  attribute INIT_10 : string;
  attribute INIT_11 : string;
  attribute INIT_12 : string;
  attribute INIT_13 : string;
  attribute INIT_14 : string;
  attribute INIT_15 : string;
  attribute INIT_16 : string;
  attribute INIT_17 : string;
  attribute INIT_18 : string;
  attribute INIT_19 : string;
  attribute INIT_1a : string;
  attribute INIT_1b : string;
  attribute INIT_1c : string;
  attribute INIT_1d : string;
  attribute INIT_1e : string;
  attribute INIT_1f : string;
  attribute INIT_20 : string;
  attribute INIT_21 : string;
  attribute INIT_22 : string;
  attribute INIT_23 : string;
  attribute INIT_24 : string;
  attribute INIT_25 : string;
  attribute INIT_26 : string;
  attribute INIT_27 : string;
  attribute INIT_28 : string;
  attribute INIT_29 : string;
  attribute INIT_2a : string;
  attribute INIT_2b : string;
  attribute INIT_2c : string;
  attribute INIT_2d : string;
  attribute INIT_2e : string;
  attribute INIT_2f : string;
  attribute INIT_30 : string;
  attribute INIT_31 : string;
  attribute INIT_32 : string;
  attribute INIT_33 : string;
  attribute INIT_34 : string;
  attribute INIT_35 : string;
  attribute INIT_36 : string;
  attribute INIT_37 : string;
  attribute INIT_38 : string;
  attribute INIT_39 : string;
  attribute INIT_3a : string;
  attribute INIT_3b : string;
  attribute INIT_3c : string;
  attribute INIT_3d : string;
  attribute INIT_3e : string;
  attribute INIT_3f : string;

  attribute INIT_00 of Board_Block : label is
   "0000000000000000000000000000000000000000000000000000000000000000";
  attribute INIT_01 of Board_Block : label is
   "0000000000000000000000000000000000000000000000000000000000000000";
  attribute INIT_02 of Board_Block : label is
   "0000450443014203441500000000000000000000000000000000000000000000";
  attribute INIT_03 of Board_Block : label is
   "4000000000000000000000000000064606460646064600000000000000000000";
  attribute INIT_04 of Board_Block : label is
   "0000000000000040004000400040000000000000000000000000400040004000";
  attribute INIT_05 of Board_Block : label is
   "0040004000400000000000000000000000004000400040004000000000000000";
  attribute INIT_06 of Board_Block : label is
   "0000000000000000000066266626662666260000000000000000000000000040";
  attribute INIT_07 of Board_Block : label is
   "0000000000000000000000000000000000000000000025642361226324650000";


  -- Standard 8x16 font taken from the Linux console font file "lat0-16.psfu"

  attribute INIT_00 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";  -- empty piece
  attribute INIT_01 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";  -- empty piece
  attribute INIT_02 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";  -- empty piece
  attribute INIT_03 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- empty piece
  attribute INIT_04 of Piece_Bitmap : label is
   "3C638E383C6186381861861800F38E0000F3EE000061C6000003800000010000";  -- the KING
  attribute INIT_05 of Piece_Bitmap : label is
   "07FFFFE007FFFFE007FFFEE00679FEF00F71CE300C73CE301E73CE381863CE18";  -- the KING
  attribute INIT_06 of Piece_Bitmap : label is
   "01E00600009FF80000C0038001FFFF800180018001FFFF8003FFFF8007FFFF80";  -- the KING
  attribute INIT_07 of Piece_Bitmap : label is
   "000000000000000000000000000000000000000000FFFE0001FFFF00018FF000";	-- the KING
  attribute INIT_08 of Piece_Bitmap : label is
   "0000800000008000000080000003E00000008000000080000000000000000000";  -- the QUEEN
  attribute INIT_09 of Piece_Bitmap : label is
   "0BFD5FE80BFD5FE803FB6FE805E6B3D0061EBC3001F2A7C0000220000001C000";  -- the QUEEN
  attribute INIT_0a of Piece_Bitmap : label is
   "004FF9000060030000FFFF80008FF88001A082C0037EBF6006FEBFA005FEBFD0";  -- the QUEEN
  attribute INIT_0b of Piece_Bitmap : label is
   "000000000000000000000000000FF800007FFF0000FFFF800087F08000700700";	-- the QUEEN
  attribute INIT_0c of Piece_Bitmap : label is
   "0001C0000003E0000003E0000003E0000001C000000000000000000000000000";	-- the BUSHOP
  attribute INIT_0d of Piece_Bitmap : label is
   "007F7E00007F7F00007C1F00007F7F00003F7E00001FFE00001FFC000007F000";	-- the BUSHOP
  attribute INIT_0e of Piece_Bitmap : label is
   "001FFC000017F40000080800000FF800000FF80000180C00003FFE00003FFE00";	-- the BUSHOP
  attribute INIT_0f of Piece_Bitmap : label is
   "00000000000000000400001007FE3FF00FFF7FF80FFFFFF80006B0000007F000";	-- the BUSHOP
  attribute INIT_10 of Piece_Bitmap : label is
   "007FF800007FE0000077000000C3000000020000000000000000000000000000";	-- the KNIGHT
  attribute INIT_11 of Piece_Bitmap : label is
   "03FF7F4003FFBE4003FFBEC003FFBEC001BFFC8001DFF30000FFE60000E78C00";	-- the KNIGHT
  attribute INIT_12 of Piece_Bitmap : label is
   "0183FF980F83FF980EC1FF901FE1FFB01BF8FF300FFEFF300FFFFF3007FFFF20";	-- the KNIGHT
  attribute INIT_13 of Piece_Bitmap : label is
   "0000000000000000001FFFF8001FFFF8001FFF98001FFF98000FFF98000FFF98";	-- the KNIGHT
  attribute INIT_14 of Piece_Bitmap : label is
   "07FFFC0007FFFC0007FFFC00079F3C00079F3C00000000000000000000000000";	-- the ROOK
  attribute INIT_15 of Piece_Bitmap : label is
   "00FFE00000FFE00000FFE00000FFE00000FFE0000180300003FFF80006000C00";	-- the ROOK
  attribute INIT_16 of Piece_Bitmap : label is
   "07FFFC0007FFFC0006000C0003FFF8000180300000FFE00000FFE00000FFE000";	-- the ROOK
  attribute INIT_17 of Piece_Bitmap : label is
   "000000000000000000000000000000001FFFFF001FFFFF001FFFFF0018000300";	-- the ROOK
  attribute INIT_18 of Piece_Bitmap : label is
   "0007F0000003F0000003E0000001C00000000000000000000000000000000000";	-- the PAWN
  attribute INIT_19 of Piece_Bitmap : label is
   "000FF800000FFC00001FFC00001FFC00001FFC00000FF8000007F0000003E000";	-- the PAWN
  attribute INIT_1a of Piece_Bitmap : label is
   "00FFFF8000FFFF8000FFFF80007FFF80007FFF00003FFE00001FFC000007F000";	-- the PAWN
  attribute INIT_1b of Piece_Bitmap : label is
   "000000000000000000000000000000000000000000FFFF8000FFFF8000FFFF80";	-- the PAWN
  attribute INIT_1c of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- not use
  attribute INIT_1d of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- not use
  attribute INIT_1e of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- not use
  attribute INIT_1f of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- not use
  attribute INIT_20 of Piece_Bitmap : label is
   "80000001800000018000000180000001800000018000000180000001ffffffff";	-- selected empty piece
  attribute INIT_21 of Piece_Bitmap : label is
   "8000000180000001800000018000000180000001800000018000000180000001";	-- selected empty piece
  attribute INIT_22 of Piece_Bitmap : label is
   "8000000180000001800000018000000180000001800000018000000180000001";	-- selected empty piece
  attribute INIT_23 of Piece_Bitmap : label is
   "ffffffff80000001800000018000000180000001800000018000000180000001";	-- selected empty piece
  attribute INIT_24 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected KING
  attribute INIT_25 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected KING
  attribute INIT_26 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected KING
  attribute INIT_27 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected KING
  attribute INIT_28 of Piece_Bitmap : label is
   "007007000087F08000FFFF80007FFF00000FF800000000000000000000000000";	-- selected QUEEN
  attribute INIT_29 of Piece_Bitmap : label is
   "05FEBFD006FEBFA0037EBF6001A082C0008FF88000FFFF8000600300004FF900";	-- selected QUEEN
  attribute INIT_2a of Piece_Bitmap : label is
   "0001C0000002200001F2A7C0061EBC3005E6B3D003FB6FE80BFD5FE80BFD5FE8";	-- selected QUEEN
  attribute INIT_2b of Piece_Bitmap : label is
   "000000000000000000008000000080000003E000000080000000800000008000";	-- selected QUEEN
  attribute INIT_2c of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected BUSHOP
  attribute INIT_2d of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected BUSHOP
  attribute INIT_2e of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected BUSHOP
  attribute INIT_2f of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected BUSHOP
  attribute INIT_30 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected KNIGHT
  attribute INIT_31 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected KNIGHT
  attribute INIT_32 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected KNIGHT
  attribute INIT_33 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected KNIGHT
  attribute INIT_34 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected ROOK
  attribute INIT_35 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected ROOK
  attribute INIT_36 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected ROOK
  attribute INIT_37 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected ROOK
  attribute INIT_38 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected PAWN
  attribute INIT_39 of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected PAWN
  attribute INIT_3a of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected PAWN
  attribute INIT_3b of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- selected PAWN
  attribute INIT_3c of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- not use
  attribute INIT_3d of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- not use
  attribute INIT_3e of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- not use
  attribute INIT_3f of Piece_Bitmap : label is
   "0000000000000000000000000000000000000000000000000000000000000000";	-- not use

begin  -- Behavioral

  -----------------------------------------------------------------------------
  --
  -- Instances of the block RAMs
  -- Each is 512 bytes, so 4K total
  --
  -----------------------------------------------------------------------------

  -- First 2.5K is used for characters (5 block RAMs)
  -- Remaining 1.5K holds the font     (3 block RAMs)
  --

  -- Port A is used for communication with the OPB
  -- Port B is for video

  Board_Block : RAMB16_S9
    port map (
      DO   => DOUTB0,
      DOP0  => DOP0,
      ADDR => CharAddr(10 downto 0),
      CLK  => Pixel_Clock,
      DI   => X"00",
      DIP0  => '1',
      EN   => '1',
      SSR  => '0',
      WE   => '0');

    Piece_Bitmap : RAMB16_S36
    port map (
      DO   => DOUTB5,
      DOP  => DOP5,
      ADDR => FontAddr(8 downto 0),
      CLK  => Pixel_Clock,
      DI   => X"00000000",
      DIP  => "1111",
      EN   => '1',
      SSR  => '0',
      WE   => '0');

  ----------------------------------------------------------------------------
  --
  -- Video controller
  --
  -----------------------------------------------------------------------------

  -- Pixel clock divider

  Pixel_clk_divider : process (CLK, RST)
  begin
    if RST = '1' then
      Pixel_Clock <= '0';
    elsif CLK'event and CLK = '1' then
      Pixel_Clock <= not Pixel_Clock;
    end if;
  end process Pixel_clk_divider;

  -- Horizontal and vertical counters

  HCounter : process (Pixel_Clock, RST)
  begin
    if RST = '1' then
      Hcount <= (others => '0');
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine = '1' then
        Hcount <= (others => '0');
      else
        Hcount <= Hcount + 1;
      end if;
    end if;
  end process HCounter;

  EndOfLine <= '1' when Hcount = HTOTAL - 1 else '0';

  VCounter: process (Pixel_Clock, RST)
  begin
    if RST = '1' then
      Vcount <= (others => '0');
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine = '1' then
        if EndOfField = '1' then
          Vcount <= (others => '0');
        else
          Vcount <= Vcount + 1;
        end if;
      end if;
    end if;
  end process VCounter;

  EndOfField <= '1' when Vcount = VTOTAL - 1 else '0';

  -- State machines to generate HSYNC, VSYNC, HBLANK, and VBLANK

  HSyncGen : process (Pixel_Clock, RST)
  begin
    if RST = '1' then
      VIDOUT_HSYNC <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine = '1' then
        VIDOUT_HSYNC <= '1';
      elsif Hcount = HSYNC - 1 then
        VIDOUT_HSYNC <= '0';
      end if;
    end if;
  end process HSyncGen;

  HBlankGen : process (Pixel_Clock, RST)
  begin
    if RST = '1' then
      HBLANK_N <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if Hcount = HSYNC + HBACK_PORCH - 1 then
        HBLANK_N <= '1';
      elsif Hcount = HSYNC + HBACK_PORCH + HACTIVE - 1 then
        HBLANK_N <= '0';
      end if;
    end if;
  end process HBlankGen;

  VSyncGen : process (Pixel_Clock, RST)
  begin
    if RST = '1' then
      VIDOUT_VSYNC <= '1';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine ='1' then
        if EndOfField = '1' then
          VIDOUT_VSYNC <= '1';
        elsif VCount = VSYNC - 1 then
          VIDOUT_VSYNC <= '0';
        end if;
      end if;
    end if;
  end process VSyncGen;

  VBlankGen : process (Pixel_Clock, RST)
  begin
    if RST = '1' then
      VBLANK_N <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if EndOfLine ='1' then
        if Vcount = VSYNC + VBACK_PORCH - 1 then
          VBLANK_N <= '1';
        elsif VCount = VSYNC + VBACK_PORCH + VACTIVE - 1 then
          VBLANK_N <= '0';
        end if;
      end if;
    end if;
  end process VBlankGen;

  -- RAM read triggers and shift register control

--  LoadChar   <= '1' when Hcount(2 downto 0) = X"5" else '0';
--  FontLoad   <= '1' when Hcount(2 downto 0) = X"6" else '0';
  LoadNShift <= '1' when Hcount(4 downto 0) = X"1F" else '0';

  -- Correction of 4 needed to calculate the character address before the
  -- character is displayed
  CharColumn <= Hcount - HSYNC - HBACK_PORCH + 19;
  Column <= CharColumn(9 downto 5);  -- / 32
  CharRow <= Vcount - VSYNC - VBACK_PORCH;
  Row <= CharRow(8 downto 5);  -- / 32

  -- Column + Row * 20
  CharAddr <= Column +
              ("000" & Row(3 downto 0) & "0000") +
              ("00000" & Row(3 downto 0) & "00");

  -- Most significant bit of character ignored
  Player <= DOUTB0(5);
  Cursor <= DOUTB0(4);
  FontAddr(8 downto 5) <= DOUTB0(3 downto 0);
  FontAddr(4 downto 0) <= CharRow(4 downto 0);

  FontData <= DOUTB5;

  -- Shift register

  ShiftRegister: process (Pixel_Clock, RST)
  begin
    if RST = '1' then
      ShiftData <= X"00000000";
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if LoadNShift = '1' then
        ShiftData <= FontData;
      else
        ShiftData <= ShiftData(30 downto 0) & ShiftData(31);
      end if;
    end if;
  end process ShiftRegister;

  VideoData <= ShiftData(31);

  Board_background: process (Column, Row)
  begin
    if (Column < 6 or Column > 13) or (Row < 4 or Row > 11) then
      Background <= '0';
    elsif Column(0) = Row(0) then
      Background <= '1';
    else
      Background <= '0';
    end if;
  end process Board_background;

  -- Video signals going to the "video DAC"

  VideoOut: process (Pixel_Clock, RST, Background, Player)
  begin
    if RST = '1' then
      RED   <= '0';
      BLUE  <= '0';
      GREEN <= '0';
    elsif Pixel_Clock'event and Pixel_Clock = '1' then
      if HBLANK_N = '1' and VBLANK_N = '1' then
        if VideoData = '1' then
          if Player = '0' then
            RED   <= '1';
            GREEN <= '0';
            BLUE  <= '0';
          else
            RED   <= '0';
            GREEN <= '0';
            BLUE  <= '1';
          end if;
        else
          if Background = '0' then
            RED   <= '0';
            GREEN <= '0';
            BLUE  <= '0';
          else
            RED   <= '1';
            GREEN <= '1';
            BLUE  <= '1';
          end if;
        end if;
      else
        RED   <= '0';
        GREEN <= '0';
        BLUE  <= '0';
      end if;
    end if;
  end process VideoOut;

  Cursor_enhance : process(Pixel_Clock, Cursor)
  begin
    if Cursor = '1' then
      VIDOUT_RED <= not RED;
      VIDOUT_GREEN <= not GREEN;
      VIDOUT_BLUE <= not BLUE;
    else
      VIDOUT_RED <= RED;
      VIDOUT_GREEN <= GREEN;
      VIDOUT_BLUE <= BLUE;
    end if;
  end process Cursor_enhance;

--  VIDOUT_CLK <= Pixel_Clock;

end Behavioral;
